/*
 * dbg_algorithms.hpp for Edge minimization in de Bruijn graphs
 * Copyright (c) 2019 Uwe Baier, Pascal Weber All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifndef DEBRUIJNGRAPH_ALGORITHMS_HPP
#define DEBRUIJNGRAPH_ALGORITHMS_HPP

#include <sdsl/csa_wt.hpp>
#include <sdsl/int_vector.hpp>

#include <deque>
#include <vector>
#include <utility>
#include <limits>

class dbg_algorithms {
	public:
		typedef sdsl::int_vector<>::size_type           size_type;

	private:
		typedef typename std::deque<std::pair<size_type,size_type>>   kmer_queue;

		//function computes the k-mer bounds of order 1
		template<class t_csa_wt_type>
		static void bounds_order_one( const t_csa_wt_type &csa, kmer_queue &Q );

		//function expects the bounds of k-mer intervals for some order k,
		// and replaces them with bounds of k-mer intervals of next higher order.
		template<class t_csa_wt_type>
		static void advance( const t_csa_wt_type &csa, kmer_queue &Q );

		//function measures the amount of edges for an edge-reduced DBG given by its kmer bounds.
		template<class t_csa_wt_type>
		static size_type measure_edges( const t_csa_wt_type &csa, const kmer_queue &Q, sdsl::bit_vector &B );

		//edge minimization algorithm
		// stop_on_min: wether algorithm should stop if minimum is certain or proceed to kmax
		// csa: fm-index
		// max_k: maximal order k to which minimization should be performed
		// B: a bitvector of size csa.size() filled with ones
		// nb_buffer: a buffer used to write the node bound indices occuring during the execution of the algorithm
		// order_edgecnt_output: function that is called for every order k with the current edgecount and the order
		template<bool stop_on_min=true,
		         class t_csa_wt_type,
		         class f_order_edgecnt_output>
		static std::pair<size_type,size_type> minimize_dbg_edges( const t_csa_wt_type &csa, size_type kmax, sdsl::bit_vector &B,
		                                                          sdsl::int_vector_buffer<> &nb_buffer,
		                                                          f_order_edgecnt_output order_edgecnt_output );

	public:		
		// function computes the number of edges of an edge-reduced DeBruijn graph of order k
		// csa: a compressed suffix array of the string using a wavelet tree and the BWT
		// k: order k from which the number of edges of the DBG shall be determined
		// B: a bitvector where the upper bounds of the k-mer intervals of order k will be marked.
		//    After function call, B has size csa.size() + 1, and B[csa.size()] will always be 1
		// function returns the number of edges from the underlying DBG.
		template<class t_csa_wt_type>
		static size_type dbg_edgecount( const t_csa_wt_type &csa, size_type k, sdsl::bit_vector &B ) {
			assert( k > 0 && k <= csa.size() );

			//initialize B
			B.resize( csa.size() + 1 );
			sdsl::util::set_to_value( B, 0 );

			//compute bounds
			kmer_queue Q;
			bounds_order_one( csa, Q );
			while (--k != 0) {
				advance( csa, Q );
			}

			//measure edges and stop
			return measure_edges( csa, Q, B );
		};

		// function computes the number of edges of an edge-reduced DeBruijn graph of order k
		// csa: a compressed suffix array of the string using a wavelet tree and the BWT
		// k: order k from which the number of edges of the DBG shall be determined
		// function returns the number of edges from the underlying DBG.
		template<class t_csa_wt_type>
		static size_type dbg_edgecount( const t_csa_wt_type &csa, size_type k ) {
			sdsl::bit_vector B;
			return dbg_edgecount( csa, k, B );
		};


		//function behaves as calling dbg_edgecount for each values from 1 to (including) k,
		// but in a more efficient manner.
		// csa: a compressed suffix array of the string using a wavelet tree and the BWT
		// k: maximal k from which the number of edges of the DBG shall be determined
		// function returns a vector of size k containing the edge count of each order k,
		// where number of edges of an order k edge-reduced graph can be queried at index (k-1).
		template<class t_csa_wt_type>
		static std::vector<size_type> dbg_edgespectrum( const t_csa_wt_type &csa, size_type k ) {
			assert( k > 0 && k <= csa.size() );

			//initialize B
			sdsl::bit_vector B( csa.size() + 1, 0 );

			//edge spectrum vector
			std::vector<size_type> ES( k );

			//send node bounds to nothing
			sdsl::int_vector_buffer<> nb_buffer("/dev/null", std::ios::out);

			//compute spectrum
			minimize_dbg_edges<false>( csa, k, B, nb_buffer, [&ES](size_type k,size_type m) { ES[k-1] = m; } );

			return ES;
		};

		// function computes the order k such that the number of edges of
		// the edge-reduced DeBruijn graph of order k are minimal under all possible values of k
		// csa: a compressed suffix array of the string using a wavelet tree and the BWT
		// B: a bitvector where the upper bounds of the k-mer intervals of the smallest DBG will be marked.
		//    After function call, B has size csa.size() + 1, and B[csa.size()] will always be 1
		// config: a config indicating where temporary files should be saved
		// function returns a pair of two integers, where the first integer corresponds to the order k whilst
		// the second integer indicates the minimal amount of edges.
		template<class t_csa_wt_type>
		static std::pair<size_type,size_type> find_min_dbg( const t_csa_wt_type &csa, sdsl::bit_vector &B, sdsl::cache_config &config ) {
			//prepare bitvector B
			//initialize B
			B.resize( csa.size() + 1 );
			sdsl::util::set_to_value( B, 0 );

			//prepare buffer with node bounds
			std::string tmp_key = sdsl::util::to_string(sdsl::util::pid())+"_"+sdsl::util::to_string(sdsl::util::id());
			std::string tmp_file_name = sdsl::cache_file_name(tmp_key, config);
			sdsl::int_vector_buffer<> nb_buffer(tmp_file_name, std::ios::out);

			//run algorithm
			auto result = minimize_dbg_edges<true>( csa, csa.size(), B, nb_buffer, [](size_type,size_type) {} );

			//retain node bounds of optimal solution
			for (auto nb : nb_buffer) B[nb] = 0;

			return result;
		};
};

///////////////////////////////////////////////////////////////////////////////

template<class t_csa_wt_type>
void dbg_algorithms::bounds_order_one( const t_csa_wt_type &csa, kmer_queue &Q ) {
	//variables needed for interval_symbols function
	typename t_csa_wt_type::size_type iv_chars;
	std::vector<unsigned char> cs ( csa.sigma );
	std::vector<size_type> rank_c_l( csa.sigma );
	std::vector<size_type> rank_c_r( csa.sigma );

	sdsl::interval_symbols( csa.wavelet_tree, 0, csa.size(), iv_chars, cs, rank_c_l, rank_c_r );
	for (size_type j = 0; j < iv_chars; j++) {
		auto c = csa.char2comp[cs[j]];
		auto l = rank_c_l[j];
		auto r = rank_c_r[j]; //right bound is exclusive!
		Q.emplace_back( csa.C[c] + l, csa.C[c] + r );
	}
}

template<class t_csa_wt_type>
void dbg_algorithms::advance( const t_csa_wt_type &csa, kmer_queue &Q ) {

	//variables needed for interval_symbols function
	typename t_csa_wt_type::size_type iv_chars;
	std::vector<unsigned char> cs ( csa.sigma );
	std::vector<size_type> rank_c_l( csa.sigma );
	std::vector<size_type> rank_c_r( csa.sigma );

	size_type n = Q.size();
	for (size_type i = 0; i < n; i++) {
		auto iv = Q.front();
		Q.pop_front();
		sdsl::interval_symbols( csa.wavelet_tree, iv.first, iv.second, iv_chars, cs, rank_c_l, rank_c_r );

		for (size_type j = 0; j < iv_chars; j++) {
			auto c = csa.char2comp[cs[j]];
			auto l = rank_c_l[j];
			auto r = rank_c_r[j]; //right bound is exclusive!
			Q.emplace_back( csa.C[c] + l, csa.C[c] + r );
		}
	}
}


template<class t_csa_wt_type>
dbg_algorithms::size_type dbg_algorithms::measure_edges( const t_csa_wt_type &csa, const kmer_queue &Q, sdsl::bit_vector &B ) {
	//step 1: mark left bounds in bitvector
	for (auto iv : Q) {
		B[iv.first] = 1;
	}
	B[csa.size()] = 1;

	//step 2: check which edges can be fused to one edge
	size_type m = csa.size();
	for (auto iv : Q) {
		auto i = iv.first;
		auto j = iv.second;

		auto is_i = csa.wavelet_tree.inverse_select( i );
		auto is_j = csa.wavelet_tree.inverse_select( j-1 );
			
		auto c_i = csa.char2comp[is_i.second];
		auto c_j = csa.char2comp[is_j.second];
		size_type lf_i = csa.C[c_i] + is_i.first;
		size_type lf_j = csa.C[c_j] + is_j.first + 1;

		//check if L[i] = ... = L[j-1] and if [LF[i],LF[j-1]] is a k-mer interval
		if (c_i == c_j && (lf_j - lf_i) == (j - i) && B[lf_i] == 1 && B[lf_j] == 1) {
			m -= (j - i - 1);
		}
	}
	return m;
}

//// EDGE MINIMIZATION ALGORITHM ////
template<bool stop_on_min,
         class t_csa_wt_type,
         class f_order_edgecnt_output>
std::pair<dbg_algorithms::size_type,dbg_algorithms::size_type> 
dbg_algorithms::minimize_dbg_edges( const t_csa_wt_type &csa, size_type kmax, sdsl::bit_vector &B,
                                    sdsl::int_vector_buffer<> &nb_buffer,
                                    f_order_edgecnt_output order_edgecnt_output ) {
	assert( csa.size() < std::numeric_limits<size_type>::max() );
	size_type n = csa.size();   // initialize
	sdsl::bit_vector F(n); 
	std::vector<std::deque<std::pair<size_type,size_type>>> Q(csa.sigma);
	std::vector<size_type> size(csa.sigma);

	B[0] = B[n] = 1;                  // boundaries of the root node
	size_type nt = csa.sigma, m = n;  // node counter and edge counter
	size_type ks = 1, ms = n;         // order with minimum number of edges ms
	size_type fusible = 0;            // inherited fusions
	size_type last = std::numeric_limits<size_type>::max();

	//variables needed for interval_symbols function
	size_type iv_chars;
	std::vector<unsigned char> cs ( csa.sigma );
	std::vector<size_type> rank_c_i( csa.sigma );
	std::vector<size_type> rank_c_j( csa.sigma );

	sdsl::interval_symbols( csa.wavelet_tree, 0, n, iv_chars, cs, rank_c_i, rank_c_j );

	for (size_type c_i = 0; c_i < iv_chars; c_i++) {
		auto c = csa.char2comp[cs[c_i]];
		auto i = csa.C[c] + rank_c_i[c_i];
		auto j = csa.C[c] + rank_c_j[c_i] - 1;
		Q[c].emplace_back(i,j);
	}

	for (size_type k = 1; k <= kmax; k++) {  
		m -= fusible;  // establish new inherited fusions
		fusible = 0;
		for (size_type c = 0; c < csa.sigma; c++) {
			for (auto iv : Q[c]) {
				size_type lb = iv.first, rb = iv.second;
				if (B[rb + 1] == 0) {
					B[rb + 1] = 1;
					nb_buffer.push_back( rb + 1 ); // EXTERNAL WRITE
					if (last == std::numeric_limits<size_type>::max()) {
						last = lb;
					}
				} else {
					if (F[rb] == 1) { 
						m += (rb - last); 
						F[rb] = 0;
					}
					last = std::numeric_limits<size_type>::max();
				}
			}
		}
		
		for (size_type c = 0; c < csa.sigma; c++) {
			size[c] = Q[c].size();
		}

		for (size_type c = 0; c < csa.sigma; c++) {  // in alphabetical order
			for (size_type q = 0; q < size[c]; q++) {
				auto iv = Q[c].front();
				Q[c].pop_front();
				size_type lb = iv.first, rb = iv.second;

				sdsl::interval_symbols( csa.wavelet_tree, lb, rb + 1, iv_chars, cs, rank_c_i, rank_c_j );
				if (iv_chars == 1) {  // reclassify edges
					auto c = csa.char2comp[cs[0]];
					auto i = csa.C[c] + rank_c_i[0];
					auto j = csa.C[c] + rank_c_j[0] - 1;
					if (B[i] == 1 && B[j+1] == 1) {  // node has no siblings
						m -= (rb - lb);
					} else {
						fusible += (rb - lb);  // possible fusion in next order
					}
					F[rb] = 1; 
				}
				for (size_type c_i = 0; c_i < iv_chars; c_i++) {
					auto c = csa.char2comp[cs[c_i]];
					auto i = csa.C[c] + rank_c_i[c_i];
					auto j = csa.C[c] + rank_c_j[c_i] - 1 ;
					if (B[i] == 0 || B[j+1] == 0) {
						if (B[j+1] == 0) {
							nt++;
							if (stop_on_min && nt >= ms) {
								return std::make_pair( ks, ms );
							}
						}
						Q[c].emplace_back(i,j);
					}
				}
			}
		}
		order_edgecnt_output( k, m );
		if (m < ms) {
			ks = k;
			ms = m;
			nb_buffer.reset(); //EXTERNAL CLEAR
		}
	}
	return std::make_pair( ks, ms );
}

#endif
